#ifndef _AFFIX_HXX_
#define _AFFIX_HXX_

#include "atypes.hxx"
#include "baseaffix.hxx"
#include "affixmgr.hxx"

/* A Prefix Entry  */

class PfxEntry : public AffEntry
{
       AffixMgr*    pmyMgr;

       PfxEntry * next;
       PfxEntry * nexteq;
       PfxEntry * nextne;

public:

  PfxEntry(AffixMgr* pmgr, affentry* dp );
  ~PfxEntry();

  struct hentry *      check(const char * word, int len, int prefix_in_cpms, 
                            const FLAG needflag = FLAG_NULL);

  struct hentry *      check_twosfx(const char * word, int len, int prefix_in_cpms, const FLAG needflag = NULL);

  char *      check_morph(const char * word, int len, int prefix_in_cpms,
                            const FLAG needflag = FLAG_NULL);

  char *      check_twosfx_morph(const char * word, int len,
                  int prefix_in_cpms, const FLAG needflag = FLAG_NULL);

  inline FLAG getFlag()   { return achar;   }
  inline const char *  getKey()    { return appnd;  } 

#ifdef HUNMORPH
  inline const char *  getMorph()    { return morphcode;  } 
#endif
  inline const char *  getCont()    { return contclass;  } 

  inline PfxEntry *    getNext()   { return next;   }
  inline PfxEntry *    getNextNE() { return nextne; }
  inline PfxEntry *    getNextEQ() { return nexteq; }

  inline void   setNext(PfxEntry * ptr)   { next = ptr;   }
  inline void   setNextNE(PfxEntry * ptr) { nextne = ptr; }
  inline void   setNextEQ(PfxEntry * ptr) { nexteq = ptr; }
};




/* A Suffix Entry */

class SfxEntry : public AffEntry
{
       AffixMgr*    pmyMgr;
       char *       rappnd;

       SfxEntry *   next;
       SfxEntry *   nexteq;
       SfxEntry *   nextne;

public:

  SfxEntry(AffixMgr* pmgr, affentry* dp );
  ~SfxEntry();

  struct hentry *   check(const char * word, int len, int optflags, 
                    AffEntry* ppfx, char ** wlst, int maxSug, int * ns,
                    const FLAG cclass = FLAG_NULL, const FLAG needflag = FLAG_NULL);

  struct hentry *   check_twosfx(const char * word, int len, int optflags, AffEntry* ppfx, const FLAG needflag = NULL);

#ifdef HUNMORPH
  char *      check_twosfx_morph(const char * word, int len, int optflags,
                 AffEntry* ppfx, const FLAG needflag = FLAG_NULL);
  struct hentry * get_next_homonym(struct hentry * he);
  struct hentry * get_next_homonym(struct hentry * word, int optflags, AffEntry* ppfx, 
    const FLAG cclass, const FLAG needflag);
#endif

  inline FLAG getFlag()   { return achar;   }
  inline const char *  getKey()    { return rappnd; } 

#ifdef HUNMORPH
  inline const char *  getMorph()    { return morphcode;  } 
#endif
  inline const char *  getCont()    { return contclass;  } 
  inline const char *  getAffix()    { return appnd; } 

  inline short getKeyLen() { return appndl; } 

  inline SfxEntry *    getNext()   { return next;   }
  inline SfxEntry *    getNextNE() { return nextne; }
  inline SfxEntry *    getNextEQ() { return nexteq; }

  inline void   setNext(SfxEntry * ptr)   { next = ptr;   }
  inline void   setNextNE(SfxEntry * ptr) { nextne = ptr; }
  inline void   setNextEQ(SfxEntry * ptr) { nexteq = ptr; }

};


#endif


