#ifndef _LANGID_HXX_
#define _LANGID_HXX_

#include "license.readme"

#include "dictmgr.hxx"
#include "myspell.hxx"

#define BOOT_WORD 50     // language pre-recognition based on BOOT_WORD sampled words
#define SAMPLE_WORD 150  // sample size = SAMPLE_WORD words
#define SAMPLE_LANG 2    // max. used languages after booting (0 = all languages)
                         // latin-2 (latin-1) characters
#define LETTERS "qwertzuiopasdfghjklyxcvbnmQWERTZUIOPASDFGHJKLYXCVBNM-"
#define LANG_LIMIT 0.2   // simple output limit for identify_data function

#define PATHLEN 1024

//#define WINDOWS // Windows environment
//#define LOG // LOG

#ifdef WINDOWS
#define PATHSEP '\\'
#define gettext
#else
#define PATHSEP '/'
#include <locale.h>
#include <libintl.h>
#endif

struct idententry {
  char * lang;
  int freq;
  MySpell * pMS;
};

class LangId
{
  DictMgr*      pDMgr;
  dictentry*    dicl; // list of dictionaries
  int           dicn; // number of dictionaries
  MySpell**     dics; // dictionaries
  char          binary_chars[256]; // character table for detecting binary data

public:
  LangId(const char * diclist_path, const char * dic_path);
  ~LangId();

  int is_text(char * data, int size);
  int tokenize(char ** words, char * data, int size, int maxword, char * wordchars);
  int identify(idententry** llst, char ** wlst, int wnum,
            int sample_word, int boot_word, int sample_lang);
  int identify(idententry** llst, char ** wlst, int wnum);
  int identify_data(char *** results, char * data, int size,
            int sample_word, int boot_word, int sample_lang);
  int identify_data(char *** results, char * data, int size);
  
private:
  char * get_random(char ** wlst, int wnum, int n);
};

#endif
