#include "license.readme"
#include "license_myspell.readme"

#include <cstring>
#include <cstdlib>
#include <cstdio>

#include "myspell.hxx"

#ifndef WINDOWS
using namespace std;
#endif

extern void mychomp(char * s);
extern char * mystrdup(const char *);


MySpell::MySpell(const char * affpath, const char * dpath)
{
    encoding = NULL;
    csconv = NULL;

    /* first set up the hash manager */
    pHMgr = new HashMgr(dpath);

    /* next set up the affix manager */
    /* it needs access to the hash manager lookup methods */
    pAMgr = new AffixMgr(affpath,pHMgr);

    /* get the preferred try string and the dictionary */
    /* encoding from the Affix Manager for that dictionary */
    char * try_string = pAMgr->get_try_string();
    encoding = pAMgr->get_encoding();
    csconv = get_current_cs(encoding);
    
    /* and finally set up the suggestion manager */
    pSMgr = new SuggestMgr(try_string, MAXSUGGESTION, pAMgr);

// ------- missing?  (N.L.) ---------------------------------    
    if (try_string) free(try_string); // NL
// ----------------------------------------------------------
    
}

MySpell::~MySpell()
{
    if (pSMgr) delete pSMgr;
    if (pAMgr) delete pAMgr;
    if (pHMgr) delete pHMgr;
    pSMgr = NULL;
    pAMgr = NULL;
    pHMgr = NULL;
    csconv= NULL;
    if (encoding) free(encoding);
    encoding = NULL;
}


// make a copy of src at destination while removing all leading
// blanks and removing any trailing periods after recording
// their presence with the abbreviation flag
// also since already going through character by character, 
// set the capitalization type
// return the length of the "cleaned" word

int MySpell::cleanword(char * dest, const char * src, int * pcaptype, int * pabbrev)
{ 
   unsigned char * p = (unsigned char *) dest;
   const unsigned char * q = (const unsigned char * ) src;

   // first skip over any leading blanks
   while ((*q != '\0') && (*q == ' ')) q++;
   
   // now strip off any trailing periods (recording their presence)
   *pabbrev = 0;
   int nl = strlen((const char *)q);
   while ((nl > 0) && (*(q+nl-1)=='.')) {
       nl--;
       (*pabbrev)++;
   }
   
   // if no characters are left it can't be an abbreviation and can't be capitalized
   if (nl <= 0) { 
       *pcaptype = NOCAP;
       *pabbrev = 0;
       *p = '\0';
       return 0;
   }

   // now determine the capitalization type of the first nl letters
   int ncap = 0;
   int nneutral = 0;
   int nc = 0;
   while (nl > 0) {
       nc++;
       if (csconv[(*q)].ccase) ncap++;
       if (csconv[(*q)].cupper == csconv[(*q)].clower) nneutral++;
       *p++ = *q++;
       nl--;
   }
   // remember to terminate the destination string
   *p = '\0';

   // now finally set the captype
   if (ncap == 0) {
        *pcaptype = NOCAP;
   } else if ((ncap == 1) && csconv[(unsigned char)(*dest)].ccase) {
        *pcaptype = INITCAP;
  } else if ((ncap == nc) || ((ncap + nneutral) == nc)){
        *pcaptype = ALLCAP;
  } else {
        *pcaptype = HUHCAP;
  }
  return nc;
} 
       

void MySpell::mkallcap(char * p)
{
  while (*p != '\0') { 
    *p = csconv[((unsigned char) *p)].cupper;
    p++;
  }
}


void MySpell::mkallsmall(char * p)
{
  while (*p != '\0') { 
    *p = csconv[((unsigned char) *p)].clower;
    p++;
  }
}


int MySpell::spell(const char * word)
{
  struct hentry * rv=NULL;
  char cw[MAXWORDLEN+1];
  char wspace[MAXWORDLEN+1];

  int wl = strlen(word);
  if (wl > (MAXWORDLEN - 1)) return 0;
  
  int captype = 0;
  int abbv = 0;
  wl = cleanword(cw, word, &captype, &abbv);
  if (wl == 0) return 1;

  int n = 0;
  int n2 = 0;
  int n3 = 0;
  // test numbers and 56%-kal, 45%-kal 50-50%, 11.20-kor, but
  // numbers with two, or more dots are bad (for example 2003.07.22.)
  while ((n < wl) && 
  	(((cw[n] <= '9') && (cw[n] >= '0')) || (((cw[n] == '.') || (cw[n] == ',')) && (n > 0)))) {
  	n++;
	if ((cw[n] == '.') || (cw[n] == ',')) {
		if (((n2 == 0) && (n > 3)) || 
			((n2 > 0) && ((cw[n-1] == '.') || (cw[n-1] == ',')))) break;
		n2++;
		n3 = n;
	}
  }

  // numbers with 3, or more digits (HU)
  if ((n == wl) && (n3 > 0) && (n - n3 > 3)) return 0;
  if ((n == wl) || (((cw[n]=='%') || (cw[n]=='')) && check(cw+n))) return 1;
  //  if ((cw[n]=='.') && check(cw + n + 1)) return 1;

  switch(captype) {
     case HUHCAP: 
     case NOCAP:   { 
                     rv = check(cw); 
                     if ((abbv) && !(rv)) {
		         		memcpy(wspace,cw,wl);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         rv = check(wspace);
                     }
		     break;
                   }
     case ALLCAP: {
    		    rv = check(cw);
		    	if (rv) break;
                if (abbv) {
		         		memcpy(wspace,cw,wl);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         rv = check(wspace);
						 if (rv) break;
                     }
				}
     case INITCAP: { 
                     mkallsmall(cw);
                     memcpy(wspace,cw,(wl+1));
		             mkinitcap(cw);
                     rv = check(wspace);
                     if (!rv) rv = check(cw);
                     if ((abbv) && !(rv)) {
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         rv = check(wspace);
			 if (!rv) {
		            memcpy(wspace,cw,wl);
                            *(wspace+wl) = '.';
                            *(wspace+wl+1) = '\0';
			    rv = check(wspace);
			 }
                     }
		     break;
                   }		   
  }
  
  if (rv) return 1;
  
  // compound word with dash (HU) I18n
  char * dash;
  int result = 0;
  if ((dash=(char *) strchr(cw,'-'))) {
      *dash='\0';      
      // examine 2 sides of the dash
      if (dash[1] == '\0') { // base word ending with dash
        if (spell(cw)) return 1;
      } else {
      // first word ending with dash: word-
        char r2 = *(dash + 1);
        dash[0]='-';
        dash[1]='\0';
        result = spell(cw);
        dash[1] = r2;
        dash[0]='\0';
        if (result && spell(dash+1) && ((strlen(dash+1) > 1) || (dash[1] == 'e') ||
	    ((dash[1] > '0') && (dash[1] < '9')))) return 1;
      }
      // affixed number in correct word
      if (result && (dash > cw) && (((*(dash-1)<='9') && (*(dash-1)>='0')) || (*(dash-1)>='.'))) {
         *dash='-';
	 n = 1;
	 if (*(dash - n) == '.') n++;
	 // search first not a number character to left from dash
	 while (((dash - n)>=cw) && ((*(dash - n)=='0') || (n < 3)) && (n < 6)) {
	    n++;
	 }
	 if ((dash - n) < cw) n--;
	 // numbers: valami1000000-hoz
	 // examine 100000-hoz, 10000-hoz 1000-hoz, 10-hoz,
	 // 56-hoz, 6-hoz
	 for(; n >= 1; n--) {
	    if ((*(dash - n) >= '0') && (*(dash - n) <= '9') && check(dash - n)) return 1;
	 }
      }
  }
  return 0;
}

struct hentry * MySpell::check(const char * word)
{
  struct hentry * he = NULL;
  int len;

  forbidden_compound = 0;
  prevcompound = 0;
  prevroot = NULL;
    
  if (pHMgr)
     he = pHMgr->lookup(word);
     
    // check forbidden words
  if ((he) && (he->astr) && (pAMgr) && (pAMgr->get_forbiddenword()) &&
    TESTAFF(he->astr,pAMgr->get_forbiddenword(),he->alen)) {
    	forbidden_compound = 1;
	if ((pAMgr->get_compound()) && TESTAFF(he->astr,pAMgr->get_compound(),he->alen)) {
		forbidden_compound = 2;
	}
	return NULL;
  }

  len = strlen(word);

  if ((he == NULL) && (pAMgr)) {
     // try stripping off affixes */
     he = pAMgr->affix_check(word, len);

     // try check compound word
     if (he) {
        if ((he->astr) && (pAMgr) &&
	    TESTAFF(he->astr,pAMgr->get_forbiddenword(),he->alen)
	    && (! TESTAFF(he->astr,pAMgr->get_onlyroot(),he->alen))) {
		forbidden_compound = 1;
		return NULL;
	    }
	prevroot = he->word;
     } else if (pAMgr->get_compound() != '\0') {
          he = pAMgr->compound_check(word, len, 
	                          0,0,100,0,NULL,NULL);
	  if ((!he) && (word[len-1]=='-')) {
	     char * dup = mystrdup(word);
	     dup[len-1] = '\0';
             he = pAMgr->compound_check(dup, len-1, 
	                          -5,0,100,1,NULL,NULL); // hu_mov_rule
	     free(dup);
	  }
	  if (he) {
		prevroot = he->word;
		prevcompound = 1;
	  }
     }

  }

  if (he) return he;
  
  return NULL;
}

int MySpell::suggest(char*** slst, const char * word)
{
  char cw[MAXWORDLEN+1];
  char wspace[MAXWORDLEN+1];
  if (! pSMgr) return 0;
  int wl = strlen(word);
  if (wl > (MAXWORDLEN-1)) return 0;
  int captype = 0;
  int abbv = 0;
  wl = cleanword(cw, word, &captype, &abbv);
  if (wl == 0) return 0;
  int ns = 0;
  *slst = NULL; // HU, nsug in pSMgr->suggest
  
  switch(captype) {
     case NOCAP:   { 
                     ns = pSMgr->suggest(slst, cw, ns);
                     if (ns>0) break;
		     break;
                   }

     case INITCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     ns = pSMgr->suggest(slst, wspace, ns);
                     for (int j=0; j < ns; j++)
                       mkinitcap((*slst)[j]);

// ------- missing suggestions for proper nouns (N.L.) -----
	             ns = pSMgr->suggest(slst, cw, ns);
// ----------------------------------------------------------
                     break;
		     
                   }

     case HUHCAP: { 
                     ns = pSMgr->suggest(slst, cw, ns);
                     if (ns == 0) {
                        memcpy(wspace,cw,(wl+1));
                        mkallsmall(wspace);
                        ns = pSMgr->suggest(slst, wspace, ns);
		     }
                     break;
                   }

     case ALLCAP: { 

                     /* spec. return version informations */
                     if ((strcmp(word, VERSION_KEYWORD) == 0) && !(*slst)) {
                        *slst = (char **) calloc(2, sizeof(char *));
	                if (*slst == NULL) return -1;
                        (*slst)[0] = mystrdup(HUNSPELL_VERSION);
                        (*slst)[1] = mystrdup(get_version());
                        ns = 2;
                        break;
                     }
     
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     ns = pSMgr->suggest(slst, wspace, ns);

// ------- missing suggestions for proper nouns (N.L.) -----
//                     if (ns == 0) {
		         mkinitcap(wspace);
		         ns = pSMgr->suggest(slst, wspace, ns);
//		     }
// ----------------------------------------------------------

                     for (int j=0; j < ns; j++)
                       mkallcap((*slst)[j]);
                       
                     break;
                   }
  }

  // expand suggestions with dot(s)
    if (abbv) {
	for (int j = 0; j < ns; j++) {
	    char * l = (char *) malloc(strlen((*slst)[j]) + 1 + abbv);
	    strcpy(l, (*slst)[j]);
	    strcat(l, word + strlen(word) - abbv);
	    free((*slst)[j]);
	    (*slst)[j] = l;			
	}
    }

  // replace '-' with ' '
  if (forbidden_compound == 2) {
    for (int j=0; j < ns; j++) {
	char * pos = strchr((*slst)[j],'-');
	if (pos) *pos = ' ';
    }
  }
  return ns;
}

int MySpell::suggest_auto(char*** slst, const char * word)
{
  char cw[MAXWORDLEN+1];
  char wspace[MAXWORDLEN+1];
  if (! pSMgr) return 0;
  int wl = strlen(word);
  if (wl > (MAXWORDLEN-1)) return 0;
  int captype = 0;
  int abbv = 0;
  wl = cleanword(cw, word, &captype, &abbv);
  if (wl == 0) return 0;
  int ns = 0;
  *slst = NULL; // HU, nsug in pSMgr->suggest
  
  switch(captype) {
     case NOCAP:   { 
                     ns = pSMgr->suggest_auto(slst, cw, ns);
                     if (ns>0) break;
		     break;
                   }

     case INITCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     ns = pSMgr->suggest_auto(slst, wspace, ns);
                     for (int j=0; j < ns; j++)
                       mkinitcap((*slst)[j]);

// ------- missing suggestions for proper nouns (N.L.) -----
	             ns = pSMgr->suggest_auto(slst, cw, ns);
// ----------------------------------------------------------
                     break;
		     
                   }

     case HUHCAP: { 
                     ns = pSMgr->suggest_auto(slst, cw, ns);
                     if (ns == 0) {
                        memcpy(wspace,cw,(wl+1));
                        mkallsmall(wspace);
                        ns = pSMgr->suggest_auto(slst, wspace, ns);
		     }
                     break;
                   }

     case ALLCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     ns = pSMgr->suggest_auto(slst, wspace, ns);

// ------- missing suggestions for proper nouns (N.L.) -----
//                     if (ns == 0) {
		         mkinitcap(wspace);
		         ns = pSMgr->suggest_auto(slst, wspace, ns);
//		     }
// ----------------------------------------------------------

                     for (int j=0; j < ns; j++)
                       mkallcap((*slst)[j]);
                     break;
                   }
  }

  // expand suggestions with dot(s)
    if (abbv) {
	for (int j = 0; j < ns; j++) {
	    char * l = (char *) malloc(strlen((*slst)[j]) + 1 + abbv);
	    strcpy(l, (*slst)[j]);
	    strcat(l, word + strlen(word) - abbv);
	    free((*slst)[j]);
	    (*slst)[j] = l;			
	}
    }

  // replace '-' with ' '
  if (forbidden_compound == 2) {
    for (int j=0; j < ns; j++) {
	char * pos = strchr((*slst)[j],'-');
	if (pos) *pos = ' ';
    }
  }
  return ns;
}


int MySpell::suggest_stems(char*** slst, const char * word)
{
  char cw[MAXWORDLEN+1];
  char wspace[MAXWORDLEN+1];
  if (! pSMgr) return 0;
  int wl = strlen(word);
  if (wl > (MAXWORDLEN-1)) return 0;
  int captype = 0;
  int abbv = 0;
  wl = cleanword(cw, word, &captype, &abbv);
  if (wl == 0) return 0;
  
  int ns = 0;

  *slst = NULL; // HU, nsug in pSMgr->suggest
  
  switch(captype) {
     case HUHCAP:
     case NOCAP:   { 
                     ns = pSMgr->suggest_stems(slst, cw, ns);

                     if ((abbv) && (ns == 0)) {
		         memcpy(wspace,cw,wl);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         ns = pSMgr->suggest_stems(slst, wspace, ns);
                     }

		     break;
                   }

     case INITCAP: { 

	             ns = pSMgr->suggest_stems(slst, cw, ns);

// ------- missing suggestions for proper nouns (N.L.) -----
                     if (ns == 0) {
                        memcpy(wspace,cw,(wl+1));
                        mkallsmall(wspace);
                        ns = pSMgr->suggest_stems(slst, wspace, ns);

		     }

                     if ((abbv) && (ns == 0)) {
		         memcpy(wspace,cw,wl);
                         mkallsmall(wspace);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         ns = pSMgr->suggest_stems(slst, wspace, ns);
                     }
		     
// ----------------------------------------------------------
                     break;
		     
                   }

     case ALLCAP: { 
                     ns = pSMgr->suggest_stems(slst, cw, ns);
		     if (ns != 0) break;
		     
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     ns = pSMgr->suggest_stems(slst, wspace, ns);

// ------- missing suggestions for proper nouns (N.L.) -----
                     if (ns == 0) {
		         mkinitcap(wspace);
		         ns = pSMgr->suggest_stems(slst, wspace, ns);
		     }
// ----------------------------------------------------------


                     if ((abbv) && (ns == 0)) {
		         memcpy(wspace,cw,wl);
                         mkallsmall(wspace);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         ns = pSMgr->suggest_stems(slst, wspace, ns);
                     }


                     break;
                   }
  }
  
  return ns;
}

int MySpell::suggest_pos_stems(char*** slst, const char * word)
{
  char cw[MAXWORDLEN+1];
  char wspace[MAXWORDLEN+1];
  if (! pSMgr) return 0;
  int wl = strlen(word);
  if (wl > (MAXWORDLEN-1)) return 0;
  int captype = 0;
  int abbv = 0;
  wl = cleanword(cw, word, &captype, &abbv);
  if (wl == 0) return 0;
  
  int ns = 0; // ns=0 = normalized input

  *slst = NULL; // HU, nsug in pSMgr->suggest
  
  switch(captype) {
     case HUHCAP:
     case NOCAP:   { 
                     ns = pSMgr->suggest_pos_stems(slst, cw, ns);

                     if ((abbv) && (ns == 0)) {
		         memcpy(wspace,cw,wl);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         ns = pSMgr->suggest_pos_stems(slst, wspace, ns);
                     }

		     break;
                   }

     case INITCAP: { 

	             ns = pSMgr->suggest_pos_stems(slst, cw, ns);

// ------- missing suggestions for proper nouns (N.L.) -----
                     if (ns == 0 || ((*slst)[0][0] == '#')) {
                        memcpy(wspace,cw,(wl+1));
                        mkallsmall(wspace);
                        ns = pSMgr->suggest_pos_stems(slst, wspace, ns);
		     }
		     
// ----------------------------------------------------------
                     break;
		     
                   }

     case ALLCAP: { 
                     ns = pSMgr->suggest_pos_stems(slst, cw, ns);
		     if (ns != 0) break;
		     
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     ns = pSMgr->suggest_pos_stems(slst, wspace, ns);

// ------- missing suggestions for proper nouns (N.L.) -----
                     if (ns == 0) {
		         mkinitcap(wspace);
		         ns = pSMgr->suggest_pos_stems(slst, wspace, ns);
		     }
// ----------------------------------------------------------
                     break;
                   }
  }

  return ns;
}

char * MySpell::get_dic_encoding()
{
  return encoding;
}

char * MySpell::get_wordchars()
{
  return pAMgr->get_wordchars();
}

char * MySpell::get_prevroot()
{
  return prevroot; // XXX not stateless, not for OOo
}

int MySpell::get_prevcompound()
{
  return prevcompound; // XXX not stateless, not for OOo
}

int MySpell::get_forbidden_compound()
{
  return forbidden_compound; // XXX not stateless, not for OOo
}

void MySpell::mkinitcap(char * p)
{
  if (*p != '\0') *p = csconv[((unsigned char)*p)].cupper;
}

struct cs_info * MySpell::get_csconv()
{
  return csconv;
}

int MySpell::put_word(const char * word)
{
    if (pHMgr) {
	return pHMgr->put_word(word, strlen(word), NULL, 0);
    }
    return 0;
}

int MySpell::put_word_suffix(const char * word, const char * suffix)
{
    if (pHMgr) {
	return pHMgr->put_word(word, strlen(word), suffix, strlen(suffix));
    }
    return 0;
}

int MySpell::put_word_pattern(const char * word, const char * pattern)
{
    struct hentry * he = NULL;
    if (pHMgr) {
	he = pHMgr->lookup(pattern);
	if (he) return pHMgr->put_word(word, strlen(word), he->astr, he->alen);
    }
    return 0;
}

const char * MySpell::get_version()
{
  return pAMgr->get_version();
}

#ifdef HUNMORPH
char * MySpell::morph(const char * word)
{
  char cw[MAXWORDLEN+1];
  char wspace[MAXWORDLEN+1];
  if (! pSMgr) return 0;
  int wl = strlen(word);
  if (wl > (MAXWORDLEN-1)) return 0;
  int captype = 0;
  int abbv = 0;
  wl = cleanword(cw, word, &captype, &abbv);
  if (wl == 0) return 0;

  char result[MAXLNLEN];
  char * st = NULL;
  
  *result = '\0';

  int n = 0;
  int n2 = 0;
  int n3 = 0;

  // test numbers and 56%-kal, 45%-kal 50-50%, 11.20-kor, but
  // numbers with two, or more dots are bad (for example 2003.07.22.)
  while ((n < wl) && 
  	(((cw[n] <= '9') && (cw[n] >= '0')) || (((cw[n] == '.') || (cw[n] == ',')) && (n > 0)))) {
  	n++;
	if ((cw[n] == '.') || (cw[n] == ',')) {
		if (((n2 == 0) && (n > 3)) || 
			((n2 > 0) && ((cw[n-1] == '.') || (cw[n-1] == ',')))) break;
		n2++;
		n3 = n;
	}
  }

  // numbers with 3, or more digits (HU)
  if ((n == wl) && (n3 > 0) && (n - n3 > 3)) return NULL;
  if ((n == wl) || ((n>0) && ((cw[n]=='%') || (cw[n]=='')) && check(cw+n))) {
  	strcat(result, cw);
	result[n - 1] = '\0';
  	if (n == wl) {
		st = pSMgr->suggest_morph(cw + n - 1);
		if (st) {
			strcat(result, st);
			free(st);
		}
	} else {
		char sign = cw[n];
		cw[n] = '\0';
		st = pSMgr->suggest_morph(cw + n - 1);
		if (st) {
			strcat(result, st);
			free(st);
		}
		strcat(result, "+"); // XXX SPEC. MORPHCODE
		cw[n] = sign;
		st = pSMgr->suggest_morph(cw + n);
		if (st) {
			strcat(result, st);
			free(st);
		}
	}
	return mystrdup(result);
  }
  //  if ((cw[n]=='.') && check(cw + n + 1)) return 1;
  
  switch(captype) {
     case NOCAP:   { 
                     st = pSMgr->suggest_morph(cw);
                     if (st) {
                        strcat(result, st);
                        free(st);
                     }
					 if (abbv) {
		         		memcpy(wspace,cw,wl);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         st = pSMgr->suggest_morph(wspace);
                     	 if (st) {
                            if (*result) strcat(result, "\n");
                            strcat(result, st);
                            free(st);
						 }
                     }
					 break;
                   }
     case INITCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     st = pSMgr->suggest_morph(wspace);
                     if (st) {
                        strcat(result, st);
                        free(st);
                     }					 
	                 st = pSMgr->suggest_morph(cw);
                     if (st) {
                        if (*result) strcat(result, "\n");
                        strcat(result, st);
                        free(st);
                     }
					 if (abbv) {
		         		 memcpy(wspace,cw,wl);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                     	 mkallsmall(wspace);
                         st = pSMgr->suggest_morph(wspace);
                     	 if (st) {
                            if (*result) strcat(result, "\n");
                            strcat(result, st);
                            free(st);
						 }
                     	 mkinitcap(wspace);
                         st = pSMgr->suggest_morph(wspace);
                     	 if (st) {
                            if (*result) strcat(result, "\n");
                            strcat(result, st);
                            free(st);
						 }
                     }
		     break;
                   }
     case HUHCAP: { 
                     st = pSMgr->suggest_morph(cw);
                     if (st) {
                        strcat(result, st);
                        free(st);
                     }
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     st = pSMgr->suggest_morph(wspace);
                     if (st) {
                        if (*result) strcat(result, "\n");
                        strcat(result, st);
                        free(st);
                     }		     
                     break;
                 }
     case ALLCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     st = pSMgr->suggest_morph(wspace);
                     if (st) {
                        strcat(result, st);
                        free(st);
                     }		     
                     mkallsmall(wspace);
                     st = pSMgr->suggest_morph(wspace);
                     if (st) {
                        if (*result) strcat(result, "\n");
                        strcat(result, st);
                        free(st);
                     }
		             mkinitcap(wspace);
		             st = pSMgr->suggest_morph(wspace);
                     if (st) {
                        if (*result) strcat(result, "\n");
                        strcat(result, st);
                        free(st);
                     }
					 if (abbv) {
                     	memcpy(wspace,cw,(wl+1));
                        *(wspace+wl) = '.';
                        *(wspace+wl+1) = '\0';
                        if (*result) strcat(result, "\n");
                     	st = pSMgr->suggest_morph(wspace);
                     	if (st) {
                        	strcat(result, st);
                        	free(st);
                     	}		     
                        mkallsmall(wspace);
                        st = pSMgr->suggest_morph(wspace);
                        if (st) {
                          if (*result) strcat(result, "\n");
                          strcat(result, st);
                          free(st);
                        }
		                mkinitcap(wspace);
		                st = pSMgr->suggest_morph(wspace);
                        if (st) {
                          if (*result) strcat(result, "\n");
                          strcat(result, st);
                          free(st);
                        }
					 }
                     break;
                   }
  }

  if (result && (*result)) return mystrdup(result);

  // compound word with dash (HU) I18n
  char * dash;
  int nresult = 0;
  if ((dash=(char *) strchr(cw,'-'))) {
      *dash='\0';      
      // examine 2 sides of the dash
      if (dash[1] == '\0') { // base word ending with dash
        if (spell(cw)) return pSMgr->suggest_morph(cw);
      } else if ((dash[1] == 'e') && (dash[2] == '\0')) { // XXX (HU) -e hat.
        if (spell(cw) && (spell("-e"))) {
			st = pSMgr->suggest_morph(cw);
			if (st) {
				strcat(result, st);
				free(st);
			}
			strcat(result,"+"); // XXX spec. separator in MORPHCODE
			st = pSMgr->suggest_morph("-e");
			if (st) {
				strcat(result, st);
				free(st);
			}
			return mystrdup(result);
		}
      } else {
      // first word ending with dash: word- XXX ???
        char r2 = *(dash + 1);
        dash[0]='-';
        dash[1]='\0';
        nresult = spell(cw);
        dash[1] = r2;
        dash[0]='\0';
        if (nresult && spell(dash+1) && ((strlen(dash+1) > 1) ||
	    	((dash[1] > '0') && (dash[1] < '9')))) {
			    st = morph(cw);
			    if (st) {
			        strcat(result, st);
				    free(st);
			    	strcat(result,"+"); // XXX spec. separator in MORPHCODE
			    }
			    st = morph(dash+1);
			    if (st) {
				    strcat(result, st);
				    free(st);
			    }
			    return mystrdup(result);			
			}
      }
      // affixed number in correct word
     if (nresult && (dash > cw) && (((*(dash-1)<='9') && 
	 		(*(dash-1)>='0')) || (*(dash-1)=='.'))) {
         *dash='-';
	 n = 1;
	 if (*(dash - n) == '.') n++;
	 // search first not a number character to left from dash
	 while (((dash - n)>=cw) && ((*(dash - n)=='0') || (n < 3)) && (n < 6)) {
	    n++;
	 }
	 if ((dash - n) < cw) n--;
	 // numbers: valami1000000-hoz
	 // examine 100000-hoz, 10000-hoz 1000-hoz, 10-hoz,
	 // 56-hoz, 6-hoz
	 for(; n >= 1; n--) {
	    if ((*(dash - n) >= '0') && (*(dash - n) <= '9') && check(dash - n)) {
	   	    strcat(result, cw);
		    result[dash - cw - n] = '\0';
			st = pSMgr->suggest_morph(dash - n);
			if (st) {
		        strcat(result, st);
				free(st);
			}
		    return mystrdup(result);			
	    }
	 }
     }
  }
  return NULL;
}

char * MySpell::morph_with_correction(const char * word)
{
  char cw[MAXWORDLEN+1];
  char wspace[MAXWORDLEN+1];
  if (! pSMgr) return 0;
  int wl = strlen(word);
  if (wl > (MAXWORDLEN-1)) return 0;
  int captype = 0;
  int abbv = 0;
  wl = cleanword(cw, word, &captype, &abbv);
  if (wl == 0) return 0;

  char result[MAXLNLEN];
  char * st = NULL;
  
  *result = '\0';
  
  
  switch(captype) {
     case NOCAP:   { 
                     st = pSMgr->suggest_morph_for_spelling_error(cw);
                     if (st) {
                        strcat(result, st);
                        free(st);
                     }
					 if (abbv) {
		         		memcpy(wspace,cw,wl);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                         st = pSMgr->suggest_morph_for_spelling_error(wspace);
                     	 if (st) {
                            if (*result) strcat(result, "\n");
                            strcat(result, st);
                            free(st);
						 }
                     }
					 break;
                   }
     case INITCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     st = pSMgr->suggest_morph_for_spelling_error(wspace);
                     if (st) {
                        strcat(result, st);
                        free(st);
                     }					 
	                 st = pSMgr->suggest_morph_for_spelling_error(cw);
                     if (st) {
                        if (*result) strcat(result, "\n");
                        strcat(result, st);
                        free(st);
                     }
					 if (abbv) {
		         		 memcpy(wspace,cw,wl);
                         *(wspace+wl) = '.';
                         *(wspace+wl+1) = '\0';
                     	 mkallsmall(wspace);
                         st = pSMgr->suggest_morph_for_spelling_error(wspace);
                     	 if (st) {
                            if (*result) strcat(result, "\n");
                            strcat(result, st);
                            free(st);
						 }
                     	 mkinitcap(wspace);
                         st = pSMgr->suggest_morph_for_spelling_error(wspace);
                     	 if (st) {
                            if (*result) strcat(result, "\n");
                            strcat(result, st);
                            free(st);
						 }
                     }
		     break;
                   }
     case HUHCAP: { 
                     st = pSMgr->suggest_morph_for_spelling_error(cw);
                     if (st) {
                        strcat(result, st);
                        free(st);
                     }
                     memcpy(wspace,cw,(wl+1));
                     mkallsmall(wspace);
                     st = pSMgr->suggest_morph_for_spelling_error(wspace);
                     if (st) {
                        if (*result) strcat(result, "\n");
                        strcat(result, st);
                        free(st);
                     }		     
                     break;
                 }
     case ALLCAP: { 
                     memcpy(wspace,cw,(wl+1));
                     st = pSMgr->suggest_morph_for_spelling_error(wspace);
                     if (st) {
                        strcat(result, st);
                        free(st);
                     }		     
                     mkallsmall(wspace);
                     st = pSMgr->suggest_morph_for_spelling_error(wspace);
                     if (st) {
                        if (*result) strcat(result, "\n");
                        strcat(result, st);
                        free(st);
                     }
		             mkinitcap(wspace);
		             st = pSMgr->suggest_morph_for_spelling_error(wspace);
                     if (st) {
                        if (*result) strcat(result, "\n");
                        strcat(result, st);
                        free(st);
                     }
					 if (abbv) {
                     	memcpy(wspace,cw,(wl+1));
                        *(wspace+wl) = '.';
                        *(wspace+wl+1) = '\0';
                        if (*result) strcat(result, "\n");
                     	st = pSMgr->suggest_morph_for_spelling_error(wspace);
                     	if (st) {
                        	strcat(result, st);
                        	free(st);
                     	}		     
                        mkallsmall(wspace);
                        st = pSMgr->suggest_morph_for_spelling_error(wspace);
                        if (st) {
                          if (*result) strcat(result, "\n");
                          strcat(result, st);
                          free(st);
                        }
		                mkinitcap(wspace);
		                st = pSMgr->suggest_morph_for_spelling_error(wspace);
                        if (st) {
                          if (*result) strcat(result, "\n");
                          strcat(result, st);
                          free(st);
                        }
					 }
                     break;
                   }
  }

  if (result) return mystrdup(result);
  return NULL;
}


#endif
