#include "hashmgr.hxx"
#include "affixmgr.hxx"
#include "suggestmgr.hxx"
#include "csutil.hxx"

#define NOCAP   0
#define INITCAP 1
#define ALLCAP  2
#define HUHCAP  3

#define HUNSPELL_VERSION "1.0-RC1"
#define VERSION_KEYWORD "XXXHUNSPELLXXX"

#define MAXSUGGESTION 15

#ifdef WINDOWS
#define DLLTEST2_API __declspec(dllexport)
#endif

#ifndef _MYSPELLMGR_HXX_
#define _MYSPELLMGR_HXX_

#ifdef WINDOWS
class DLLTEST2_API MySpell
#else
class MySpell
#endif
{
  AffixMgr*       pAMgr;
  HashMgr*        pHMgr;
  SuggestMgr*     pSMgr;
  char *          encoding;
  struct cs_info * csconv;

/* not stateless variables */
  char *	  prevroot;
  int             prevcompound;

/* forbidden_compound:
 * 0 = not forbidden
 * 1 = forbidden
 * 2 = forbidden compound (written without dash in Hungarian)
 */
  int		  forbidden_compound;
  
public:
  MySpell(const char * affpath, const char * dpath);
  ~MySpell();

  int suggest(char*** slst, const char * word);
  int suggest_auto(char*** slst, const char * word);
  int suggest_stems(char*** slst, const char * word);
  int suggest_pos_stems(char*** slst, const char * word);
  int spell(const char *);
  char * get_dic_encoding();
  char * get_wordchars();
  struct cs_info * get_csconv();
  char * get_prevroot();
  int get_prevcompound();
  int get_forbidden_compound();
  int put_word(const char * word);
  int put_word_suffix(const char * word, const char * suffix);
  int put_word_pattern(const char * word, const char * pattern);
  char * get_possible_root();
  const char * get_version();
  char * morph(const char * word);
  char * morph_with_correction(const char * word);
  
private:
   int    cleanword(char *, const char *, int *, int *);
   void   mkinitcap(char *);
   void   mkallcap(char *);
   void   mkallsmall(char *);
   struct hentry * check(const char *);
   void   initialize_possible_root();

};

#endif
