/*************************************************************************
 (C) 2002 Noll Janos <spdaemon@johnzero.hu>, Prim Kft.
    license: GNU GPL

 Some ideas and code taken from CMU Cyrus IMAPd/pwcheck daemon,
 (C) 1999, Carnegie Mellon University

 spdaemon - spell checker daemon
 
 The daemon resides in memory, and listens on a unix socket,
 or on a TCP port (#define-configurable)
 
 Input (text):
  hu_HU\n   <- locale code
  >\n       <- separator
  knownword1\n   <- known words
  knownword2\n
  ...
  knownwordn\n
  >\n
  line1\n    <- text lines to check
  line2\n
  \0         <- \0 must be present at the end of the input!

 Output:
  cline1\n   <- checked lines
  cline2\n

 On error:
  [not yet specified]
 
  where the checked lines:
  - <, > and | symbols are escaped to \< \> and \|
  - if a misspelled word is found, suggestions can be made:
    <badword|suggest1|suggest2|...|suggestn>
  - if there are no suggestions:
    <badword>

*************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <paths.h>
#include <syslog.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/resource.h>
#include <sys/time.h>

#include <mcheck.h>

/*******************************************************/

#define SP_TCPPORT	4201
#define SP_LISTEN_TCP	// listen on tcp - otherwise: local socket
#define SPCHECKDIR	"/var/www/spcheck"

// #define SPCHECKDIR	"/tmp"
//#define STRESSTEST
#define SP_DODAEMON	// run as daemon

/*******************************************************/

#if !defined(_PATH_SPCHECKPID)
#ifdef _PATH_VARRUN
# define _PATH_SPCHECKPID (_PATH_VARRUN "spcheck.pid")
#else
# define _PATH_SPCHECKPID (NULL)
#endif
#endif


int spcheck_init( char *datadir, char *lcodes );
char *spcheck_process( char *reqstart, char *reqend );
int spcheck_end( void );

void newclient(int);
int retry_write(int, const char *, unsigned int);

int start_tcp_port( void ) {
    struct sockaddr_in srvaddr;
    int s,r;

    memset((char *)&srvaddr, 0, sizeof(srvaddr));
    srvaddr.sin_family = AF_INET;
    srvaddr.sin_addr.s_addr = 0;    /*  0.0.0.0  ==  this host, all int.  */
    bzero(srvaddr.sin_zero, 8);
    srvaddr.sin_port = htons(SP_TCPPORT);
    s = socket(AF_INET, SOCK_STREAM,0);
    if (s == -1) {
	perror("socket");
	exit(1);
    }
    r = bind(s, (struct sockaddr *)&srvaddr, sizeof(srvaddr) );
    if (r == -1) {
	syslog(LOG_ERR, "bind error: ");
	exit(1);
    }

    return s;
}

int start_domain( void ) {
    char fnamebuf[MAXPATHLEN];
    struct sockaddr_un srvaddr;
    mode_t oldumask;
    int s,r;

    s = socket(AF_UNIX, SOCK_STREAM, 0);
    if (s == -1) {
	perror("socket");
	exit(1);
    }

    strncpy(fnamebuf, SPCHECKDIR, sizeof(fnamebuf));
    strncpy(fnamebuf + sizeof(SPCHECKDIR)-1, "/spcheck",
	    sizeof(fnamebuf) - sizeof(SPCHECKDIR));
    fnamebuf[MAXPATHLEN-1] = '\0';

    (void) unlink(fnamebuf);

    memset((char *)&srvaddr, 0, sizeof(srvaddr));
    srvaddr.sun_family = AF_UNIX;
    strncpy(srvaddr.sun_path, fnamebuf, sizeof(srvaddr.sun_path));
    /* Most systems make sockets 0777 no matter what you ask for.
       Known exceptions are Linux and DUX. */
    oldumask = umask((mode_t) 0); /* for Linux, which observes the umask when
			    setting up the socket */
    r = bind(s, (struct sockaddr *)&srvaddr, sizeof(srvaddr));
    if (r == -1) {
	syslog(LOG_ERR, "%.*s: ",
	       sizeof(srvaddr.sun_path), srvaddr.sun_path);
	exit(1);
    }
    umask(oldumask); /* for Linux */
    chmod(fnamebuf, (mode_t) 0777); /* for DUX, where this isn't the default.
				    (harmlessly fails on some systems) */	

    return s;
}

/********************************************************/

int main( int argc, char **argv )
{
    int c,s;
    int count;
    int rc;
    struct sockaddr_un clientaddr;
    int r;
    socklen_t len;
    char *pid_file = _PATH_SPCHECKPID;
    FILE *fp = NULL;
    pid_t pid;

    openlog("spcheck", LOG_NDELAY, LOG_AUTH);

#ifndef STRESSTEST
    /* normal operation */
    
    if (argc<2) {
	printf(
	    "Usage: spdaemon <datadir> <localecode1[,localecode2[,...]]>>\n"
	    "Example: spdaemon /usr/share/dicts hu_HU,en_US\n"
	);
	exit(0);
    }

    spcheck_init(argv[1],argv[2]); /* init spcheck */

#else
    /* if test run, no daemon mode */
	
    void *mp1,*mp2,*mp3,*mp4;

    mp1=sbrk(0);

    spcheck_init(".","hu_HU"); /* init spcheck */

    mp2=mp4=sbrk(0);
    printf("mem1: %ld\n",(char *)mp2-(char *)mp1);

    rusage ru;


    char rb[3000];
    char *reply;

    //mtrace();
    for (c=0;c<5000;c++) {
    //for (c=0;c<3;c++) {
	strcpy(rb,"hu_HU\nMSSS\n>\nEz egy MSSS proba mondat\njol mondom\nezt");
	reply = spcheck_process(rb, rb + strlen(rb) + 1);
        //printf("R: %s\n",reply);

        //mp3=sbrk(0);
        //printf("mem3: %ld, %ld\n",(char *)mp3-(char *)mp2,(char *)mp3-(char *)mp4);
	//mp4=mp3;
    }

    mp3=sbrk(0);
    printf("memend: %ld\n",(char *)mp3-(char *)mp2);

    spcheck_end();

    exit(0);
#endif
    
#ifdef SP_DODAEMON
    /* Daemonize. */
    count = 5;
    while (count--) {
	pid = fork();
            
	if (pid > 0)
	    _exit(0);               /* parent dies */
            
	if ((pid == -1) && (errno == EAGAIN)) {
	    syslog(LOG_WARNING, "master fork failed (sleeping): ");
	    sleep(5);
	    continue;
	}
    }
    if (pid == -1) {
	rc = errno;
	syslog(LOG_ERR, "FATAL: master fork failed: ");
	fprintf(stderr, "spcheck: ");
	errno = rc;
	perror("fork");
	exit(1);
    }

    /*
     * We're now running in the child. Lose our controlling terminal
     * and obtain a new process group.
     */
    if (setsid() == -1) {
	rc = errno;
	syslog(LOG_ERR, "FATAL: setsid: ");
	fprintf(stderr, "spcheck: ");
	errno = rc;
	perror("setsid");
	exit(1);
    }
        
    s = open("/dev/null", O_RDWR, 0);
    if (s == -1) {
	rc = errno;
	syslog(LOG_ERR, "FATAL: /dev/null: ");
	fprintf(stderr, "spcheck: ");
	errno = rc;
	perror("/dev/null");
	exit(1);
            
    }
    dup2(s, fileno(stdin));
    dup2(s, fileno(stdout));
    dup2(s, fileno(stderr));
    if (s > 2) {
	close(s);
    }
#endif

    /*
     *   Record process ID - shamelessly stolen from inetd (I.V.)
     */
    pid = getpid();
    if (pid_file) {
	fp = fopen(pid_file, "w");
    }
    if (fp) {
        fprintf(fp, "%ld\n", (long)pid);
        fclose(fp);
    } else if (pid_file) {
        syslog(LOG_WARNING, "%s: ", pid_file);
    }

#ifdef SP_LISTEN_TCP
    s = start_tcp_port();
#else
    s = start_domain();
#endif

    r = listen(s, 5);
    if (r == -1) {
	syslog(LOG_ERR, "listen: ");
	exit(1);
    }

    for (;;) {
	len = sizeof(clientaddr);
	c = accept(s, (struct sockaddr *)&clientaddr, &len);
	if (c == -1 && errno != EINTR) {
	    syslog(LOG_WARNING, "accept: ");
	    continue;
	}

	newclient(c);
    }
}

void newclient(int c)
{
    char request[64000];
    int n;
    unsigned int start;
    char *reply;
    extern char *pwcheck();
    
    start = 0;
    while (start < sizeof(request) - 1) {
	n = read(c, request+start, sizeof(request) - 1 - start);
	if (n < 1) {
	    reply = "Error reading request";
	    goto sendreply;
	}
		
	start += n;

	if (request[start-1] == '\0' && strlen(request) < start) {
	    break;
	}
    }

    if (start >= sizeof(request) - 1) {
	reply = "Request too big";
    }
    else {
	reply = spcheck_process(request, request + strlen(request) + 1);
    }

sendreply:

    retry_write(c, reply, strlen(reply));

    close(c);
}
  
/*
 * Keep calling the write() system call with 'fd', 'buf', and 'nbyte'
 * until all the data is written out or an error occurs.
 */
int retry_write(int fd, const char *buf, unsigned int nbyte)
{
    int n;
    int written = 0;

    if (nbyte == 0)
	return 0;

    for (;;) {
        n = write(fd, buf, nbyte);
        if (n == -1) {
            if (errno == EINTR)
		continue;
            return -1;
        }

        written += n;

        if ((unsigned int) n >= nbyte)
	    return written;

        buf += n;
        nbyte -= n;
    }
}
